@@ -1,6 +1,7 @@
 /* sdlevent.c --- SDL input handling for Guile
  *
- * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2009, 2011 Thien-Thi Nguyen
+ * Copyright (C) 2003, 2004, 2005, 2006, 2007,
+ *   2009, 2011, 2013 Thien-Thi Nguyen
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -19,210 +20,43 @@
  */
 
 #include "guile-sdl.h"
-#include <SDL/SDL_events.h>
+#include "SDL_events.h"
+#include "SDL_active.h"
+#include "b-values.h"
 
 /* enum/flag types */
+
+static SCM active_enum;
+#include "k/active.c"
+
+static SCM updn_enum;
+#include "k/updn.c"
+
+static SCM mbut_enum;
+#include "k/mbut.c"
+
 static SCM event_type_enum;
-static valaka_t event_type_eback[] = {
-  VALAKA (SDL_ACTIVEEVENT),
-  VALAKA (SDL_KEYDOWN),
-  VALAKA (SDL_KEYUP),
-  VALAKA (SDL_MOUSEMOTION),
-  VALAKA (SDL_MOUSEBUTTONDOWN),
-  VALAKA (SDL_MOUSEBUTTONUP),
-  VALAKA (SDL_JOYAXISMOTION),
-  VALAKA (SDL_JOYBALLMOTION),
-  VALAKA (SDL_JOYHATMOTION),
-  VALAKA (SDL_JOYBUTTONDOWN),
-  VALAKA (SDL_JOYBUTTONUP),
-  VALAKA (SDL_QUIT),
-  VALAKA (SDL_SYSWMEVENT),
-  VALAKA (SDL_VIDEORESIZE),
-  VALAKA (SDL_USEREVENT)
-};
+#include "k/evtype.c"
 
 static SCM event_state_enum;
-static valaka_t event_state_eback[] = {
-  VALAKA (SDL_QUERY),
-  VALAKA (SDL_IGNORE),
-  /* SDL_DISABLE is not mentioned in the associated comment in SDL_events.h
-     (SDL 1.2), and moreover, its value is the same as SDL_IGNORE, so we
-     tickle the irony bone a bit and don't include it in Guile-SDL.  */
-  /* VALAKA (SDL_DISABLE), */
-  VALAKA (SDL_ENABLE)
-};
+#include "k/evstate.c"
 
 static SCM event_keysym_enum;
-static valaka_t event_keysym_eback[] = {
-  VALAKA (SDLK_BACKSPACE),
-  VALAKA (SDLK_TAB),
-  VALAKA (SDLK_CLEAR),
-  VALAKA (SDLK_RETURN),
-  VALAKA (SDLK_PAUSE),
-  VALAKA (SDLK_ESCAPE),
-  VALAKA (SDLK_SPACE),
-  VALAKA (SDLK_EXCLAIM),
-  VALAKA (SDLK_QUOTEDBL),
-  VALAKA (SDLK_HASH),
-  VALAKA (SDLK_DOLLAR),
-  VALAKA (SDLK_AMPERSAND),
-  VALAKA (SDLK_QUOTE),
-  VALAKA (SDLK_LEFTPAREN),
-  VALAKA (SDLK_RIGHTPAREN),
-  VALAKA (SDLK_ASTERISK),
-  VALAKA (SDLK_PLUS),
-  VALAKA (SDLK_COMMA),
-  VALAKA (SDLK_MINUS),
-  VALAKA (SDLK_PERIOD),
-  VALAKA (SDLK_SLASH),
-  VALAKA (SDLK_0),
-  VALAKA (SDLK_1),
-  VALAKA (SDLK_2),
-  VALAKA (SDLK_3),
-  VALAKA (SDLK_4),
-  VALAKA (SDLK_5),
-  VALAKA (SDLK_6),
-  VALAKA (SDLK_7),
-  VALAKA (SDLK_8),
-  VALAKA (SDLK_9),
-  VALAKA (SDLK_COLON),
-  VALAKA (SDLK_SEMICOLON),
-  VALAKA (SDLK_LESS),
-  VALAKA (SDLK_EQUALS),
-  VALAKA (SDLK_GREATER),
-  VALAKA (SDLK_QUESTION),
-  VALAKA (SDLK_AT),
-  VALAKA (SDLK_LEFTBRACKET),
-  VALAKA (SDLK_BACKSLASH),
-  VALAKA (SDLK_RIGHTBRACKET),
-  VALAKA (SDLK_CARET),
-  VALAKA (SDLK_UNDERSCORE),
-  VALAKA (SDLK_BACKQUOTE),
-  VALAKA (SDLK_a),
-  VALAKA (SDLK_b),
-  VALAKA (SDLK_c),
-  VALAKA (SDLK_d),
-  VALAKA (SDLK_e),
-  VALAKA (SDLK_f),
-  VALAKA (SDLK_g),
-  VALAKA (SDLK_h),
-  VALAKA (SDLK_i),
-  VALAKA (SDLK_j),
-  VALAKA (SDLK_k),
-  VALAKA (SDLK_l),
-  VALAKA (SDLK_m),
-  VALAKA (SDLK_n),
-  VALAKA (SDLK_o),
-  VALAKA (SDLK_p),
-  VALAKA (SDLK_q),
-  VALAKA (SDLK_r),
-  VALAKA (SDLK_s),
-  VALAKA (SDLK_t),
-  VALAKA (SDLK_u),
-  VALAKA (SDLK_v),
-  VALAKA (SDLK_w),
-  VALAKA (SDLK_x),
-  VALAKA (SDLK_y),
-  VALAKA (SDLK_z),
-  VALAKA (SDLK_DELETE),
-  VALAKA (SDLK_KP0),
-  VALAKA (SDLK_KP1),
-  VALAKA (SDLK_KP2),
-  VALAKA (SDLK_KP3),
-  VALAKA (SDLK_KP4),
-  VALAKA (SDLK_KP5),
-  VALAKA (SDLK_KP6),
-  VALAKA (SDLK_KP7),
-  VALAKA (SDLK_KP8),
-  VALAKA (SDLK_KP9),
-  VALAKA (SDLK_KP_PERIOD),
-  VALAKA (SDLK_KP_DIVIDE),
-  VALAKA (SDLK_KP_MULTIPLY),
-  VALAKA (SDLK_KP_MINUS),
-  VALAKA (SDLK_KP_PLUS),
-  VALAKA (SDLK_KP_ENTER),
-  VALAKA (SDLK_KP_EQUALS),
-  VALAKA (SDLK_UP),
-  VALAKA (SDLK_DOWN),
-  VALAKA (SDLK_RIGHT),
-  VALAKA (SDLK_LEFT),
-  VALAKA (SDLK_INSERT),
-  VALAKA (SDLK_HOME),
-  VALAKA (SDLK_END),
-  VALAKA (SDLK_PAGEUP),
-  VALAKA (SDLK_PAGEDOWN),
-  VALAKA (SDLK_F1),
-  VALAKA (SDLK_F2),
-  VALAKA (SDLK_F3),
-  VALAKA (SDLK_F4),
-  VALAKA (SDLK_F5),
-  VALAKA (SDLK_F6),
-  VALAKA (SDLK_F7),
-  VALAKA (SDLK_F8),
-  VALAKA (SDLK_F9),
-  VALAKA (SDLK_F10),
-  VALAKA (SDLK_F11),
-  VALAKA (SDLK_F12),
-  VALAKA (SDLK_F13),
-  VALAKA (SDLK_F14),
-  VALAKA (SDLK_F15),
-  VALAKA (SDLK_NUMLOCK),
-  VALAKA (SDLK_CAPSLOCK),
-  VALAKA (SDLK_SCROLLOCK),
-  VALAKA (SDLK_RSHIFT),
-  VALAKA (SDLK_LSHIFT),
-  VALAKA (SDLK_RCTRL),
-  VALAKA (SDLK_LCTRL),
-  VALAKA (SDLK_RALT),
-  VALAKA (SDLK_LALT),
-  VALAKA (SDLK_RMETA),
-  VALAKA (SDLK_LMETA),
-  VALAKA (SDLK_LSUPER),
-  VALAKA (SDLK_RSUPER),
-  VALAKA (SDLK_MODE),
-  VALAKA (SDLK_HELP),
-  VALAKA (SDLK_PRINT),
-  VALAKA (SDLK_SYSREQ),
-  VALAKA (SDLK_BREAK),
-  VALAKA (SDLK_MENU),
-  VALAKA (SDLK_POWER),
-  VALAKA (SDLK_EURO)
-};
+#include "k/keysym.c"
 
 static SCM event_action_enum;
-static valaka_t event_action_eback[] = {
-  VALAKA (SDL_ADDEVENT),
-  VALAKA (SDL_PEEKEVENT),
-  VALAKA (SDL_GETEVENT)
-};
+#include "k/evaction.c"
 
+static SCM event_mb_flags;
 static SCM event_mod_flags;
 static SCM event_mask_flags;
+static SCM event_jhpos_flags;
+static SCM appstate_flags;
 
 #define ALLOCA_EVENTS(count)  alloca (count * sizeof (SDL_Event))
 
-PRIMPROC
-(get_event_mod_flags, "flagstash:event-mod", 0, 0, 0,
- (void),
- doc: /***********
-Return the flagstash object for event mod flags.
-@xref{Enums and Constants}.  */)
-{
-  return event_mod_flags;
-}
-
-PRIMPROC
-(get_event_mask_flags, "flagstash:event-mask", 0, 0, 0,
- (void),
- doc: /***********
-Return the flagstash object for event mask flags.
-@xref{Enums and Constants}.  */)
-{
-  return event_mask_flags;
-}
-
 
-static long event_tag;
+static smob_tag_t event_tag;
 
 #define event_nick "SDL-Event"
 
@@ -250,30 +84,6 @@ free_event (SCM event)
   return GCRV (cevent);
 }
 
-
-static long keysym_tag;
-
-#define keysym_nick "SDL-Keysym"
-
-#define ASSERT_KEYSYM(obj,which) \
-  ASSERT_SMOB (obj, keysym, which)
-
-#define UNPACK_KEYSYM(smob) \
-  (SMOBGET (smob, SDL_keysym *))
-
-#define RETURN_NEW_KEYSYM(x) \
-  NEWSMOB_OR_FALSE (keysym_tag, x)
-
-static
-size_t
-free_keysym (SCM keysym)
-{
-  SDL_keysym *ckeysym = UNPACK_KEYSYM (keysym);
-
-  GCFREE (ckeysym, keysym_nick);
-  return GCRV (ckeysym);
-}
-
 
 /* Constructors */
 
@@ -284,17 +94,16 @@ PRIMPROC
  (SCM type),
  doc: /***********
 Return a new SDL event.
-Optional arg @var{type} is one of the symbols
-enumerated in the variable @code{event-types}.
-If omitted, the default is @code{SDL_NOEVENT}.
-@xref{Enums and Constants}.  */)
+Optional arg @var{type} is a symbol (@pxref{event-type enums}).
+If omitted, the default is @code{SDL_NOEVENT}.  */)
 {
 #define FUNC_NAME s_make_event
+  DECLINIT_SYM2NUM_CC (1, event_type_enum);
   SDL_Event *event;
   int ctype = SDL_NOEVENT;
 
   if (BOUNDP (type))
-    ctype = GSDL_ENUM2LONG (type, event_type_enum, 1);
+    ctype = ENUM2LONG (1, type);
 
   if ((event = GCMALLOC_EVENT ()))
     event->type = ctype;
@@ -303,41 +112,6 @@ If omitted, the default is @code{SDL_NOEVENT}.
 #undef FUNC_NAME
 }
 
-PRIMPROC
-(make_keysym, "make-keysym", 0, 2, 0,
- (SCM sym, SCM mod),
- doc: /***********
-Return a new keysym.  Optional args @var{sym} and @var{mod}
-specify one of the @code{event-keys} (@pxref{Enums and Constants})
-and any modifiers (from @code{flasgstash:event-mod}), respectively.  */)
-{
-#define FUNC_NAME s_make_keysym
-  SDL_keysym *keysym;
-
-  if ((keysym = GCMALLOC (sizeof (SDL_keysym), keysym_nick)))
-    {
-      /* Set the sym if given.  */
-      UNBOUND_MEANS_FALSE (sym);
-      if (NOT_FALSEP (sym))
-        {
-          ASSERT_INTEGER (sym, 1);
-          /* keysym->sym = (SDLKey) C_LONG (sym); */
-          keysym->sym = (SDLKey) GSDL_ENUM2LONG (sym, event_keysym_enum, 1);
-        }
-
-      /* Set the mod if given.  */
-      if (BOUNDP (mod))
-        {
-          ASSERT_INTEGER (mod, 2);
-          keysym->mod = (SDLMod) GSDL_FLAGS2ULONG (mod, event_mod_flags, 2);
-        }
-    }
-
-  /* Return the new smob.  */
-  RETURN_NEW_KEYSYM (keysym);
-#undef FUNC_NAME
-}
-
 
 /* Smob getters and setters */
 
@@ -364,7 +138,13 @@ and any modifiers (from @code{flasgstash:event-mod}), respectively.  */)
                       event_ ## c_frag,                 \
                       event, Event, c_field)
 
-#define NUMBER_SETTER(s_frag, c_frag, c_field)  \
+#define SNUMBER_SETTER(s_frag, c_frag, c_field) \
+  GSDL_NUMBER_SETTER ("event:" s_frag,          \
+                      event_ ## c_frag,         \
+                      event, Event,             \
+                      c_field, C_LONG)
+
+#define UNUMBER_SETTER(s_frag, c_frag, c_field) \
   GSDL_NUMBER_SETTER ("event:" s_frag,          \
                       event_ ## c_frag,         \
                       event, Event,             \
@@ -375,14 +155,23 @@ and any modifiers (from @code{flasgstash:event-mod}), respectively.  */)
                     a ## _ ## b,                \
                         a.b)
 
-#define NUM2_SETTER(a      ,      b)            \
-  NUMBER_SETTER   (#a   ":set-"  #b "!",        \
-                    a ## _set_ ## b,            \
-                          a.b)
+#define SNUM2_SETTER(a      ,      b)           \
+  SNUMBER_SETTER   (#a   ":set-"  #b "!",       \
+                     a ## _set_ ## b,           \
+                           a.b)
 
-#define NUM2_GETSET(a, b) \
+#define UNUM2_SETTER(a      ,      b)           \
+  UNUMBER_SETTER   (#a   ":set-"  #b "!",       \
+                     a ## _set_ ## b,           \
+                           a.b)
+
+#define SNUM2_GETSET(a, b) \
   NUM2_GETTER (a, b)      \
-  NUM2_SETTER (a, b)
+  SNUM2_SETTER (a, b)
+
+#define UNUM2_GETSET(a, b) \
+  NUM2_GETTER (a, b)      \
+  UNUM2_SETTER (a, b)
 
 
 #define NUM3_GETTER(a    ,    b    ,    c)      \
@@ -390,14 +179,23 @@ and any modifiers (from @code{flasgstash:event-mod}), respectively.  */)
                     a ## _ ## b ## _ ## c,      \
                             a.b.c)
 
-#define NUM3_SETTER(a    ,    b      ,      c)          \
-  NUMBER_SETTER   (#a   ":"  #b   ":set-"  #c "!",      \
+#define SNUM3_SETTER(a    ,    b      ,      c)         \
+  SNUMBER_SETTER   (#a   ":"  #b   ":set-"  #c "!",     \
                     a ## _ ## b ## _set_ ## c,          \
                             a.b.c)
 
-#define NUM3_GETSET(a, b, c) \
-  NUM3_GETTER (a, b, c)      \
-  NUM3_SETTER (a, b, c)
+#define UNUM3_SETTER(a    ,    b      ,      c)         \
+  UNUMBER_SETTER   (#a   ":"  #b   ":set-"  #c "!",     \
+                    a ## _ ## b ## _set_ ## c,          \
+                            a.b.c)
+
+#define SNUM3_GETSET(a, b, c) \
+  NUM3_GETTER (a, b, c)       \
+  SNUM3_SETTER (a, b, c)
+
+#define UNUM3_GETSET(a, b, c) \
+  NUM3_GETTER (a, b, c)       \
+  UNUM3_SETTER (a, b, c)
 
 
 #define FLAG_GETTER(s_frag, c_frag, c_field, stash)     \
@@ -423,9 +221,19 @@ ENUM_GETSET ("type",          type,
              type,
              event_type)
 
-NUM2_GETSET (active, gain)
-NUM2_GETSET (active, state)
-NUM2_GETSET (key, state)
+ENUM_GETSET ("active:gain",        active_gain,
+             "active:set-gain!", active_set_gain,
+             active.gain,
+             active)
+FLAG_GETSET ("active:state",        active_state,
+             "active:set-state!", active_set_state,
+             active.state,
+             appstate_flags)
+
+ENUM_GETSET ("key:state",       key_state,
+             "key:set-state!", key_set_state,
+             key.state,
+             updn)
 
 ENUM_GETSET ("key:keysym:sym",        key_keysym_sym,
              "key:keysym:set-sym!", key_keysym_set_sym,
@@ -437,39 +245,54 @@ FLAG_GETSET ("key:keysym:mod",        key_keysym_mod,
              key.keysym.mod,
              event_mod_flags)
 
-NUM3_GETSET (key, keysym, scancode)
-NUM3_GETSET (key, keysym, unicode)
+UNUM3_GETSET (key, keysym, scancode)
+UNUM3_GETSET (key, keysym, unicode)
 
-NUM2_GETSET (motion, state)
-NUM2_GETSET (motion, x)
-NUM2_GETSET (motion, y)
-NUM2_GETSET (motion, xrel)
-NUM2_GETSET (motion, yrel)
+FLAG_GETSET ("motion:state",        motion_state,
+             "motion:set-state!", motion_set_state,
+             motion.state,
+             event_mb_flags)
+UNUM2_GETSET (motion, x)
+UNUM2_GETSET (motion, y)
+SNUM2_GETSET (motion, xrel)
+SNUM2_GETSET (motion, yrel)
 
-NUM2_GETSET (button, button)
-NUM2_GETSET (button, state)
-NUM2_GETSET (button, x)
-NUM2_GETSET (button, y)
+ENUM_GETSET ("button:button",        button_button,
+             "button:set-button!", button_set_button,
+             button.button,
+             mbut)
+ENUM_GETSET ("button:state",         button_state,
+             "button:set-state!", button_set_state,
+             button.state,
+             updn)
+UNUM2_GETSET (button, x)
+UNUM2_GETSET (button, y)
 
-NUM2_GETSET (jaxis, which)
-NUM2_GETSET (jaxis, axis)
-NUM2_GETSET (jaxis, value)
+UNUM2_GETSET (jaxis, which)
+UNUM2_GETSET (jaxis, axis)
+SNUM2_GETSET (jaxis, value)
 
-NUM2_GETSET (jbutton, which)
-NUM2_GETSET (jbutton, button)
-NUM2_GETSET (jbutton, state)
+UNUM2_GETSET (jbutton, which)
+UNUM2_GETSET (jbutton, button)
+ENUM_GETSET ("jbutton:state",        jbutton_state,
+             "jbutton:set-state!", jbutton_set_state,
+             jbutton.state,
+             updn)
 
-NUM2_GETSET (jball, which)
-NUM2_GETSET (jball, ball)
-NUM2_GETSET (jball, xrel)
-NUM2_GETSET (jball, yrel)
+UNUM2_GETSET (jball, which)
+UNUM2_GETSET (jball, ball)
+SNUM2_GETSET (jball, xrel)
+SNUM2_GETSET (jball, yrel)
 
-NUM2_GETSET (jhat, which)
-NUM2_GETSET (jhat, hat)
-NUM2_GETSET (jhat, value)
+UNUM2_GETSET (jhat, which)
+UNUM2_GETSET (jhat, hat)
+FLAG_GETSET ("jhat:value",        jhat_value,
+             "jhat:set-value!", jhat_set_value,
+             jhat.value,
+             event_jhpos_flags)
 
-NUM2_GETSET (resize, w)
-NUM2_GETSET (resize, h)
+SNUM2_GETSET (resize, w)
+SNUM2_GETSET (resize, h)
 
 #if 0 /* what is this? --ttn */
 NUM2_GETSET (user, code)
@@ -492,11 +315,135 @@ Gather events from input devices and update the event queue.  */)
 #undef FUNC_NAME
 }
 
+PRIMPROC
+(evqueue_add, "evqueue-add", 0, 0, 1,
+ (SCM events),
+ doc: /***********
+Add @code{events} to the back of the event queue.
+Return the count of succesfully added events.  */)
+{
+#define FUNC_NAME s_evqueue_add
+  SDL_Event *cevents = NULL;
+  int i, count;
+  SCM ls, head;
+
+  /* Pass 1: Validate argument types and consistency.  */
+  SCM_VALIDATE_LIST_COPYLEN (1, events, count);
+  for (i = 0, ls = events;
+       i < count;
+       i++, ls = CDR (ls))
+    {
+      head = CAR (ls);
+      ASSERT_EVENT (head, 1 + i);
+    }
+
+  /* Pass 2: Allocate the array and copy the events (ugh).  */
+  if (! (cevents = ALLOCA_EVENTS (count)))
+    abort ();
+  for (i = 0, ls = events;
+       i < count;
+       i++, ls = CDR (ls))
+    cevents[i] = *(UNPACK_EVENT (CAR (ls)));
+
+  RETURN_INT
+    (SDL_PeepEvents (cevents, count, SDL_ADDEVENT, 0));
+#undef FUNC_NAME
+}
+
+struct evqueue_do_details {
+  const char            *who;
+  const SDL_eventaction  act;
+  const bool             acc;
+};
+
+static SCM
+evqueue_do (const struct evqueue_do_details *dd, SCM n, SCM mask)
+{
+  const char *FUNC_NAME = dd->who;
+  DECLINIT_SYM2NUM_CC (2, event_mask_flags);
+  SDL_Event *got;
+  unsigned long cn, cmask;
+  int count;
+
+  ASSERT_ULONG_COPY (n, 1);
+  cmask = FLAGS2ULONG (2, mask);
+
+  if (! (got = ALLOCA_EVENTS (cn))
+      || 0 > (count = SDL_PeepEvents (got, cn, dd->act, cmask)))
+    return BOOL_FALSE;
+
+  if (! dd->acc)
+    RETURN_INT (count);
+  else
+    {
+      SCM ls = SCM_EOL;
+
+      while (count--)
+        {
+          SDL_Event *cev = GCMALLOC_EVENT ();
+          SCM ev;
+
+          *cev = got[count];
+          NEW_EVENT_X (ev, cev);
+          ls = CONS (ev, ls);
+        }
+      return ls;
+    }
+}
+
+PRIMPROC
+(evqueue_peek, "evqueue-peek", 2, 1, 0,
+ (SCM n, SCM mask, SCM accumulate),
+ doc: /***********
+Return a count (less than or equal to @var{n}) of events at
+the front of the event queue that match @var{mask},
+without changing the queue.  Optional arg @var{accumulate} if
+non-@code{#f} means to return the list of matched events, instead.
+If there are errors, return @code{#f}.
+
+@xref{event-mask flags}.  */)
+{
+  const struct evqueue_do_details dd =
+    {
+      .who = s_evqueue_peek,
+      .act = SDL_PEEKEVENT,
+      .acc = BOUNDP (accumulate) && NOT_FALSEP (accumulate)
+    };
+
+  return evqueue_do (&dd, n, mask);
+}
+
+PRIMPROC
+(evqueue_get, "evqueue-get", 2, 0, 0,
+ (SCM n, SCM mask),
+ doc: /***********
+Return a list (of length at most @var{n}) of
+events at the front of the event queue that match
+@var{mask}, removing them from the queue.
+If there are errors, return @code{#f}.
+
+@xref{event-mask flags}.  */)
+{
+  const struct evqueue_do_details dd =
+    {
+      .who = s_evqueue_get,
+      .act = SDL_GETEVENT,
+      .acc = true
+    };
+
+  return evqueue_do (&dd, n, mask);
+}
+
 PRIMPROC
 (peep_events, "peep-events", 4, 0, 0,
  (SCM events, SCM numevents, SCM action, SCM mask),
  doc: /***********
-Manage the event queue, depending on @var{action}:
+NB: This procedure is obsoleted by @code{evqueue-add},
+@code{evqueue-peek} and @code{evqueue-get};
+it @strong{will be removed} after 2013-12-31.
+
+Manage the event queue, depending
+on @var{action} (@pxref{event-action enums}):
 
 @table @code
 @item SDL_ADDEVENT
@@ -506,7 +453,7 @@ Add up to @var{numevents} (an integer) events from
 @item SDL_PEEKEVENT
 Return a count (number less than or equal to @var{numevents})
 of events at the front of the event queue that match @var{mask}
-(see @code{flagstash:event-mask}), without changing the queue.
+(@pxref{event-mask flags}), without changing the queue.
 
 @item SDL_GETEVENT
 Act like for @code{SDL_PEEKEVENT} except return a list of
@@ -514,6 +461,8 @@ matching events instead of a count, removing them from the queue.
 @end table  */)
 {
 #define FUNC_NAME s_peep_events
+  DECLINIT_SYM2NUM_CC (3, event_action_enum);
+  DECLINIT_SYM2NUM_CC (4, event_mask_flags);
   SDL_Event *cevents = NULL;
   SDL_eventaction caction;
   int cnumevents, i, ret = -1;
@@ -521,7 +470,7 @@ matching events instead of a count, removing them from the queue.
   SCM ls = SCM_BOOL_F;
 
   ASSERT_LONG_COPY (numevents, 2);
-  caction = GSDL_ENUM2LONG (action, event_action_enum, 3);
+  caction = ENUM2LONG (3, action);
 
   switch (caction)
     {
@@ -552,7 +501,7 @@ matching events instead of a count, removing them from the queue.
       /* fallthrough */
 
     case SDL_PEEKEVENT:
-      cmask = GSDL_FLAGS2ULONG (mask, event_mask_flags, 4);
+      cmask = FLAGS2ULONG (4, mask);
       ret = SDL_PeepEvents (cevents, cnumevents, caction, cmask);
       if (0 > ret)
         SCM_MISC_ERROR ("badness", SCM_EOL);
@@ -646,16 +595,14 @@ PRIMPROC
 (push_event, "push-event", 1, 0, 0,
  (SCM event),
  doc: /***********
-Push @var{event} onto the queue.  Return 1 for success,
-0 if the queue was full, -1 for other errors.  */)
+Push @var{event} onto the queue.
+Return @code{#t} on success.  */)
 {
 #define FUNC_NAME s_push_event
-  int result;
-
   ASSERT_EVENT (event, 1);
 
-  result = SDL_PushEvent (UNPACK_EVENT (event));
-  RETURN_INT (result);
+  RETURN_TRUE_IF_0
+    (SDL_PushEvent (UNPACK_EVENT (event)));
 #undef FUNC_NAME
 }
 
@@ -739,30 +686,47 @@ an event type, or @code{#t} if the proc takes an event object.  */)
 }
 
 PRIMPROC
-(event_state, "event-state", 2, 0, 0,
- (SCM type, SCM state),
+(event_type_handling, "event-type-handling", 1, 1, 0,
+ (SCM type, SCM setting),
  doc: /***********
-Query or set event @var{type} to @var{state}.
-@var{type} should be one elements from @code{event-types},
-and likewise @var{state} from @code{event-states}.
-@xref{Enums and Constants}.
-If @var{state} is @code{SDL_QUERY}, return the current
-processing state of the specified event.  */)
+Return @code{#t} if event @var{type} (@pxref{event-type enums})
+is recognized and queued, or @code{#f} if it is ignored.
+If @var{setting} is specified, set the handling of
+@var{type} to the truth value of @var{setting} first.  */)
 {
-#define FUNC_NAME s_event_state
-  int ctype, cstate, ret;
+#define FUNC_NAME s_event_type_handling
+  DECLINIT_SYM2NUM_CC   (1, event_type_enum);
+  int ctype = ENUM2LONG (1, type);
 
-  ctype = GSDL_ENUM2LONG (type, event_type_enum, 1);
-  cstate = GSDL_ENUM2LONG (state, event_state_enum, 2);
-
-  ret = SDL_EventState (ctype, cstate);
-  if (SDL_QUERY == cstate)
-    return btw->long2enum (ret, event_state_enum);
-  else
-    RETURN_UNSPECIFIED;
+  return BOOLEAN (SDL_ENABLE == SDL_EventState
+                  (ctype, CSTATE_FROM_SETTING (setting)));
 #undef FUNC_NAME
 }
 
+PRIMPROC
+(event_state, "event-state", 2, 0, 0,
+ (SCM type, SCM state),
+ doc: /***********
+NB: This procedure is obsoleted by @code{event-type-handling}
+and @strong{will be removed} after 2013-12-31.
+
+Set or query the state of event @var{type} (@pxref{event-type enums})
+processing, based on @var{state} (@pxref{event-state enums}).
+If @var{state} is @code{SDL_QUERY}, return the current state.
+If it is @code{SDL_IGNORE} or @code{SDL_ENABLE},
+disable or enable, respectively, internal event @var{type}
+processing and return @var{state}.  */)
+{
+#define FUNC_NAME s_event_state
+  DECLINIT_SYM2NUM_CC   (1, event_type_enum);
+  int ctype = ENUM2LONG (1, type);
+  DECLINIT_SYM2NUM_CC    (2, event_state_enum);
+  int cstate = ENUM2LONG (2, state);
+  int ret = SDL_EventState (ctype, cstate);
+
+  return btw->long2enum (ret, event_state_enum);
+#undef FUNC_NAME
+}
 
 PRIMPROC
 (enable_unicode, "enable-unicode", 0, 1, 0,
@@ -809,7 +773,7 @@ PRIMPROC
 (get_key_state, "get-key-state", 0, 0, 0,
  (),
  doc: /***********
-Return a list of pressed keys (SDLK_* symbols).  */)
+Return a list of pressed keys (@pxref{keysym enums}).  */)
 {
 #define FUNC_NAME s_get_key_state
   Uint8 *keystate;
@@ -846,67 +810,169 @@ a list of symbols.  This does not change the keyboard state,
 only the key modifier flags.  */)
 {
 #define FUNC_NAME s_set_mod_state
-  ASSERT_INTEGER (modstate, 1);
-  SDL_SetModState (GSDL_FLAGS2ULONG (modstate, event_mod_flags, 1));
+  DECLINIT_SYM2NUM_CC          (1, event_mod_flags);
+  SDL_SetModState (FLAGS2ULONG (1, modstate));
   RETURN_UNSPECIFIED;
 #undef FUNC_NAME
 }
 
+typedef Uint8 SDLCALL (*getmouse_sdl_fn) (int *x, int *y);
+
 DECLARE_SIMPLE_SYM (state);
 DECLARE_SIMPLE_SYM (x);
 DECLARE_SIMPLE_SYM (y);
 
-PRIMPROC
-(get_mouse_state, "get-mouse-state", 0, 0, 0,
- (void),
- doc: /***********
-Return the current state of the mouse as an alist with
-symbolic keys: @code{state}, @code{x} and @code{y}.  */)
+static SCM
+getmouse (const char *FUNC_NAME, SCM symbolic, getmouse_sdl_fn fn)
 {
-#define FUNC_NAME s_get_mouse_state
-  int buttons, x, y;
-  buttons = SDL_GetMouseState (&x, &y);
-  RETURN_LIST3 (CONS (SYM (state), NUM_LONG (buttons)),
-                CONS (SYM (x), NUM_LONG (x)),
-                CONS (SYM (y), NUM_LONG (y)));
-#undef FUNC_NAME
+  unsigned long buttons;
+  int x, y;
+
+  UNBOUND_MEANS_FALSE (symbolic);
+
+  buttons = fn (&x, &y);
+  return LIST3
+    (CONS (SYM (state),
+           EXACTLY_FALSEP (symbolic)
+           ? NUM_ULONG (buttons)
+           : (buttons
+              ? btw->ulong2flags (buttons, event_mb_flags)
+              : SCM_EOL)),
+     CONS (SYM (x), NUM_LONG (x)),
+     CONS (SYM (y), NUM_LONG (y)));
 }
 
 PRIMPROC
-(get_relative_mouse_state, "get-mouse-relative-state", 0, 0, 0,
- (void),
+(get_mouse_state, "get-mouse-state", 0, 1, 0,
+ (SCM symbolic),
  doc: /***********
-Return the current relative state of the mouse as an alist
-symbolic keys: @code{state}, @code{x} and @code{y}.  */)
+NB: This procedure is obsoleted by @code{mouse-bxy}
+and @strong{will be removed} after 2013-12-31.
+
+Return the current state of the mouse as an alist with
+symbolic keys: @code{state}, @code{x} and @code{y}.
+Normally, all values are integers.
+However, if optional arg @var{symbolic} is non-@code{#f},
+the @code{state} value is instead a (possibly empty)
+list of symbols of the set:
+
+@example
+left middle right
+wheel-up wheel-down
+x1 x2
+@end example  */)
 {
-#define FUNC_NAME s_get_relative_mouse_state
-  int buttons, x, y;
-  buttons = SDL_GetRelativeMouseState (&x, &y);
-  RETURN_LIST3 (CONS (SYM (state), NUM_LONG (buttons)),
-                CONS (SYM (x), NUM_LONG (x)),
-                CONS (SYM (y), NUM_LONG (y)));
-#undef FUNC_NAME
+  return getmouse (s_get_mouse_state,
+                   symbolic,
+                   SDL_GetMouseState);
+}
+
+PRIMPROC
+(get_relative_mouse_state, "get-mouse-relative-state", 0, 1, 0,
+ (SCM symbolic),
+ doc: /***********
+NB: This procedure is obsoleted by @code{mouse-bxy}
+and @strong{will be removed} after 2013-12-31.
+
+Return the current relative state of the mouse as an alist
+with symbolic keys: @code{state}, @code{x} and @code{y}.
+Optional arg @var{symbolic} has the same effect as
+for @code{get-mouse-state}.  */)
+{
+  return getmouse (s_get_relative_mouse_state,
+                   symbolic,
+                   SDL_GetRelativeMouseState);
 }
 
 PRIMPROC
 (button_p, "button?", 1, 0, 0,
  (SCM mask),
  doc: /***********
-Return @code{#t} if buttons specified in @var{mask} (an integer)
-are pressed.  Use 1 for left, 2 for middle and 4 for right,
-combined with @code{logior}, to form @var{mask}.  For example,
-a value of 5 specifies both left and right buttons.  */)
+Return @code{#t} if buttons specified in @var{mask} are pressed,
+otherwise @code{#f}.
+@var{mask} is a symbol or a list of symbols from the set returned
+by @code{get-mouse-state}.
+
+For backward compatability, @var{mask} can also be the (integer)
+logior of the buttons, using mapping:
+
+@example
+ 1  left
+ 2  middle
+ 4  right
+ 8  wheel-up
+16  wheel-down
+32  x1
+64  x2
+@end example
+
+For example, a value of 5 specifies both left and right buttons,
+equivalent to @code{(left right)}.  */)
 {
 #define FUNC_NAME s_button_p
-  ASSERT_INTEGER (mask, 1);
-  RETURN_BOOL
-    (SDL_BUTTON (C_LONG (mask)));
+  unsigned long cmask, buttons;
+
+  if (INTEGERP (mask))
+    ASSERT_ULONG_COPY (mask, 1);
+  else
+    {
+      DECLINIT_SYM2NUM_CC (1, event_mb_flags);
+      cmask = FLAGS2ULONG (1, mask);
+    }
+
+  if (! cmask)
+    return BOOL_FALSE;
+
+  buttons = SDL_GetMouseState (NULL, NULL);
+  RETURN_BOOL (cmask == (cmask & buttons));
 #undef FUNC_NAME
 }
 
+PRIMPROC
+(mouse_bxy, "mouse-bxy", 0, 1, 0,
+ (SCM relative),
+ doc: /***********
+Return three values: a (possibly empty) list of symbols
+representing pressed mouse buttons (like @code{event:button:button}),
+and two integer coordinates @var{x} and @var{y}.
+
+Optional arg @code{relative} non-@code{#f} means the
+coordinates are relative to the last time the underlying
+@code{SDL_GetRelativeMouseState} was called.  */)
+{
+#define FUNC_NAME s_mouse_bxy
+  getmouse_sdl_fn fn;
+  int buttons, x, y;
+
+  UNBOUND_MEANS_FALSE (relative);
+  fn = EXACTLY_FALSEP (relative)
+    ? SDL_GetMouseState
+    : SDL_GetRelativeMouseState;
+
+  buttons = fn (&x, &y);
+  RETURN_VALUES3
+    (btw->ulong2flags (buttons, event_mb_flags),
+     NUM_INT (x),
+     NUM_INT (y));
+#undef FUNC_NAME
+}
+
+PRIMPROC
+(get_app_state, "get-app-state", 0, 0, 0,
+ (void),
+ doc: /***********
+Return the current state of the application, a list of symbols.
+The list may include: `mousefocus', `inputfocus', `active'.  */)
+{
+  return btw->ulong2flags (SDL_GetAppState (), appstate_flags);
+}
+
 
-#include "kmod.c"
-#include "evmask.c"
+#include "k/mb.c"
+#include "k/kmod.c"
+#include "k/jhpos.c"
+#include "k/evmask.c"
+#include "k/appstate.c"
 
 /* Initialize glue.  */
 void
@@ -917,22 +983,34 @@ gsdl_init_event (void)
            free_event,
            /* TODO: print_event */ NULL);
 
-  DEFSMOB (keysym_tag, keysym_nick,
-           NULL,
-           free_keysym,
-           /* TODO: print_keysym */ NULL);
+  /* constants */
+  {
+    kp_init_t allp[] = {
+      { &event_type_enum, &evtype_kp },
+      { &event_keysym_enum, &keysym_kp },
+      { &event_action_enum, &evaction_kp },
+      { &event_state_enum, &evstate_kp },
+      { &mbut_enum, &mbut_kp },
+      { &updn_enum, &updn_kp },
+      { &active_enum, &active_kp }
+    };
 
-  /* event type constants */
-  event_type_enum = DEFINE_ENUM ("event-types", event_type_eback);
+    REGISTER_KP_V (allp);
+  }
+  {
+    kf_init_t allf[] = {
+      { &event_mb_flags, &mb_flagstash },
+      { &event_mod_flags, &kmod_flagstash },
+      { &event_mask_flags, &evmask_flagstash },
+      { &event_jhpos_flags, &jhpos_flagstash },
+      { &appstate_flags, &appstate_flagstash }
+    };
 
-  /* keysyms */
-  event_keysym_enum = DEFINE_ENUM ("event-keys", event_keysym_eback);
-  event_action_enum = DEFINE_ENUM ("event-actions", event_action_eback);
-  event_mod_flags = btw->make_flagstash (&kmod_flagstash);
-  event_mask_flags = btw->make_flagstash (&evmask_flagstash);
+    REGISTER_KF_V (allf);
+  }
 
-  /* event states */
-  event_state_enum = DEFINE_ENUM ("event-states", event_state_eback);
+  btw->event_state_enum = event_state_enum;
+  btw->updn_enum = updn_enum;
 
   efi.proc = SCM_BOOL_F;
 
