@@ -21,7 +21,7 @@
 
 #define GUILE_SDL_OPTIONAL_MODULE  1
 #include "guile-sdl.h"
-#include <SDL/SDL_ttf.h>
+#include "SDL_ttf.h"
 #include "snuggle/finangle.h"
 #include "b-values.h"
 
@@ -32,7 +32,7 @@ SELECT_MODULE_VAR (obtw, sdlsup, "%%Guile-SDL-obtw");
 static SCM ttf_flags;
 
 /* Smob tags.  */
-static long ttf_font_tag;
+static smob_tag_t ttf_font_tag;
 #define ttf_font_nick "SDL-TTF"
 
 #define ASSERT_TTFONT(obj,which) \
